/*

Copyright 2018,2019,2021 Surface Concept GmbH, Germany

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.

*/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <thread>
#include <chrono>
#include <time.h>
#include "sema.h"

#include <scTDC.h>
#include <scTDC_error_codes.h>


using namespace std;

class usercallbacks_example {
  int device_descriptor = 0;
  int nr_measurements = 10;
  int measurement_counter = 1;
  int acqtime_ms = 100;
  bool measurements_finished = false;
  bool abort_program = false;
  double binsize_time = 0.5;
  std::ofstream* outf_ptr = nullptr; // pointer to the text file stream object
  Semaphore semaphore;

  static void _s_on_end_of_measure(void *);
  static void _s_on_dld_event(void *, const struct sc_DldEvent *const, size_t);
  static void _s_on_start_of_measurement(void *);
  static void _s_on_millisecond_countup(void *);
  static void _s_on_statistics(void *, const struct statistics_t *);
  static void _s_on_tdc_event(void *, const struct sc_TdcEvent *const, size_t);
  void on_end_of_measure();
  void on_dld_event(const struct sc_DldEvent *const, size_t);
  void on_tdc_event(const struct sc_TdcEvent *const, size_t);
  void on_start_of_measurement();
  void on_millisecond_countup();
  void on_statistics(const struct statistics_t *);
  std::string end_of_measurement_message();
  int start_measurement();

public:
  usercallbacks_example();

  int write_tdc_events_to_textfile(std::ofstream& outf,
                                   long acqtime_ms_arg,
                                   long nrmeas_arg);
};

// constructor implementation
usercallbacks_example::usercallbacks_example()
{

}

// ---------------- BOILERPLATE ------------------------------------------------
// can only register static "member" functions of classes to the USER_CALLBACKS
// pipe. Use private pointer to call regular member functions of the class
// object.
void usercallbacks_example::_s_on_end_of_measure(void *p) {
  reinterpret_cast<usercallbacks_example*>(p)->on_end_of_measure();
}

void usercallbacks_example::_s_on_start_of_measurement(void *p) {
  reinterpret_cast<usercallbacks_example*>(p)->on_start_of_measurement();
}

void usercallbacks_example::_s_on_millisecond_countup(void *p) {
  reinterpret_cast<usercallbacks_example*>(p)->on_millisecond_countup();
}

void usercallbacks_example::_s_on_dld_event(void *p,
  const struct sc_DldEvent *const e, size_t s)
{
  reinterpret_cast<usercallbacks_example*>(p)->on_dld_event(e, s);
}

void usercallbacks_example::_s_on_tdc_event(void *p,
  const struct sc_TdcEvent *const e, size_t s)
{
  reinterpret_cast<usercallbacks_example*>(p)->on_tdc_event(e, s);
}

void usercallbacks_example::_s_on_statistics(void *p,
  const struct statistics_t *s)
{
  reinterpret_cast<usercallbacks_example*>(p)->on_statistics(s);
}
// --------- END OF BOILERPLATE ------------------------------------------------

void usercallbacks_example::on_end_of_measure() 
{
  // FILL IN USERCODE
  if (measurement_counter < nr_measurements) 
  {
    measurement_counter++;
    // ---------------------------------------------
    // ---    signalize main thread to wake up   ---
    // ---------------------------------------------
    semaphore.signal();
  }
  else
  {
    measurements_finished = true;
    semaphore.signal();
  }
}

void usercallbacks_example::on_start_of_measurement() 
{
  // FILL IN USERCODE
  std::ofstream& outf = *(outf_ptr);
  outf << "# next measurement" << std::endl;
}

void usercallbacks_example::on_millisecond_countup() 
{
  // FILL IN USERCODE
}

void usercallbacks_example::on_tdc_event(
  const struct sc_TdcEvent *const e, size_t s)
{
  // FILL IN USERCODE
  if (outf_ptr == nullptr) // safety check
    return;
  std::ofstream& outf = *(outf_ptr);

  // output values of the event data fields into file (one row per event)
  std::ostringstream oss;
  for (size_t i = 0; i < s; i++) {
    oss << std::setw(8) << e[i].channel
      << std::setw(8) << e[i].subdevice
      << std::setw(25) << std::setprecision(
           std::numeric_limits<double>::digits10 + 1)
      << (static_cast<double>(e[i].time_data) * binsize_time * 1e-6)
         // time (1e-6 converts from ns to ms)
      << std::setw(20) << e[i].start_counter
      << std::setw(20) << e[i].sign_counter
      << std::setw(20) << e[i].time_tag
      << std::endl;
  }
  outf << oss.str();
}

void usercallbacks_example::on_dld_event(
  const struct sc_DldEvent *const e, size_t event_array_len)
{
  // FILL IN USERCODE
  (void)e; // does nothing, prevents compiler warning
  (void)event_array_len; // does nothing, prevents compiler warning
}

void usercallbacks_example::on_statistics(const struct statistics_t *stat)
{
  // FILL IN USERCODE
  (void) stat;  // does nothing, prevents compiler warning
}


std::string usercallbacks_example::end_of_measurement_message() 
{
  std::ostringstream oss;
  std::chrono::system_clock::time_point tnow_tmp =
      std::chrono::system_clock::now();
  auto duration = tnow_tmp.time_since_epoch();
  auto millis =
      std::chrono::duration_cast<std::chrono::milliseconds>(duration).count();
  millis = millis % 1000;
  time_t tnow_tmp2 = std::chrono::system_clock::to_time_t(tnow_tmp);
  struct tm tnow;
#if defined(__WIN32__) || defined(_WIN32)
  localtime_s(&tnow, &tnow_tmp2);
#else
  localtime_r(&tnow_tmp2, &tnow);
#endif
  oss << "# END OF MEASUREMENT   DATE & TIME : " <<
    std::setw(4) << std::setfill('0') << tnow.tm_year + 1900 << "-" <<
    std::setw(2) << std::setfill('0') << tnow.tm_mon + 1 << "-" <<
    std::setw(2) << std::setfill('0') << tnow.tm_mday << " " <<
    std::setw(2) << std::setfill('0') << tnow.tm_hour << ":" <<
    std::setw(2) << std::setfill('0') << tnow.tm_min << ":" <<
    std::setw(2) << std::setfill('0') << tnow.tm_sec << " (+" <<
    std::setw(3) << std::setfill('0') << millis << "ms)" << std::endl
      << std::endl;
  return oss.str();
}

int usercallbacks_example::start_measurement()
{
  std::cout << "Starting measurement #" 
            << std::setw(4) << measurement_counter
            << std::endl;
  int retcode;
  // sometimes starting a measurement directly after the previous
  // fails with NOT_READY error. Retry two times.
  for (unsigned i = 0; i < 3; i++)
  {
     retcode = sc_tdc_start_measure2(device_descriptor, acqtime_ms);
     if (retcode != SC_TDC_ERR_NOTRDY) break;
     std::this_thread::sleep_for(std::chrono::milliseconds(1));
  }
  if (retcode < 0) 
  {
    std::cout << "sc_tdc_start_measure2 returned " << retcode 
              << std::endl;
    char errmsg[ERRSTRLEN];
    sc_get_err_msg(retcode, errmsg);
    std::cout << errmsg << std::endl;
    abort_program = true;
  }
  return retcode;
}

int usercallbacks_example::write_tdc_events_to_textfile(
    std::ofstream &outf, long acqtime_ms_arg, long nrmeas_arg)
{
  // ---------------------------------------------------------------
  // ---               initialize device                         ---
  // ---------------------------------------------------------------
  int retcode = sc_tdc_init_inifile("tdc_gpx3.ini");
  if (retcode < 0) {
    std::cout << "Init Inifile returned " << retcode << std::endl;
    char errmsg[ERRSTRLEN];
    sc_get_err_msg(retcode, errmsg);
    std::cout << errmsg << std::endl;
    return retcode;
  }
  device_descriptor = retcode;
  acqtime_ms = acqtime_ms_arg;
  nr_measurements = nrmeas_arg;
  outf_ptr = &outf;

  // ---------------------------------------------------------------
  // ---        get time bin size in nanoseconds                 ---
  // ---------------------------------------------------------------
  sc_tdc_get_binsize2(device_descriptor, &binsize_time); // TDC time bin size
  std::cout << "# Time pixel corresponds to " << binsize_time << " ns"
            << std::endl;
  // ---------------------------------------------------------------
  // ---     open a pipe for receiving events via callbacks      ---
  // ---  1. fill a structure with function pointers to our      ---
  // ---     callback handlers                                   ---
  // ---  2. open the pipe                                       ---
  // ---------------------------------------------------------------
  sc_pipe_callback_params_t cb_params;
  cb_params.callbacks = new sc_pipe_callbacks;
  // put "this" pointer into priv: we get it back in the callbacks
  cb_params.callbacks->priv = reinterpret_cast<void*>(this);
  cb_params.callbacks->end_of_measure =
      usercallbacks_example::_s_on_end_of_measure;
  cb_params.callbacks->dld_event = usercallbacks_example::_s_on_dld_event;
  cb_params.callbacks->millisecond_countup =
      usercallbacks_example::_s_on_millisecond_countup;
  cb_params.callbacks->start_of_measure =
      usercallbacks_example::_s_on_start_of_measurement;
  cb_params.callbacks->statistics = usercallbacks_example::_s_on_statistics;
  cb_params.callbacks->tdc_event = usercallbacks_example::_s_on_tdc_event;
  sc_pipe_open2(device_descriptor, USER_CALLBACKS,
                  reinterpret_cast<void*>(&cb_params));
  // ---------------------------------------------------------------
  // ---       write header into the output file                 ---
  // ---------------------------------------------------------------
  outf << "#" << std::setw(7) << "chan" // channel
    << std::setw(8) << "subdev"
    << std::setw(25) << "time (ms)"
    << std::setw(20) << "start counter"
    << std::setw(20) << "sign counter"
    << std::setw(20) << "time tag"
    << std::endl;
  // ---------------------------------------------------------------
  // ---  loop through repeated starts of measurements           ---
  // ---------------------------------------------------------------
  while (!(measurements_finished || abort_program))
  {
    int retcode2 = start_measurement();
    if (retcode2 < 0)
    {
      sc_tdc_deinit2(device_descriptor);
      return retcode2;
    }    
    semaphore.wait(); // the "on_end_of_measure" function signalizes
                      // this thread to wake up
  }
  std::this_thread::sleep_for(std::chrono::milliseconds(100)); // wait once more
  // ---------------------------------------------------------------
  // ---              release TDC                                ---
  // ---------------------------------------------------------------
  std::cout << end_of_measurement_message();
  sc_tdc_deinit2(device_descriptor);
  return 0;
}






int main(int argc, char **argv)
{
  // ----------------------------------------------------------------
  // ---     parse command line arguments here, call              ---
  // ---     write_tdc_events_to_textfile(...) to perform what    ---
  // ---     the program is aimed at                              ---
  // ----------------------------------------------------------------
  if (argc != 4 && argc != 5) { // program name itself is counted in argc
    std::cout << std::endl << "usercallbacks_example  -  Surface Concept GmbH"
              << std::endl;
    std::cout << "Version for TDC (not DLD) 22.10.2019" << std::endl
              << std::endl;
    std::cout << "Usage: " << std::endl
      << "  usercallbacks_example.exe  "
         "[-f] OUTPUTFILE  ACQUISITION_TIME(ms)  NUMBER_OF_MEASUREMENTS"
      << std::endl
      << "                            -f ... overwrite OUTPUTFILE if exists"
      << std::endl;
    exit(1);
  }
  // at this point, argc == 4 || argc == 5
  std::size_t argshift = 0; // if optional argument is used, shift the rest
  bool overwrite = false;
  std::string arg1{ argv[1] };
  if (arg1.compare("-f") == 0) {
    overwrite = true;
    argshift++;
  }

  if (overwrite && argc != 5) {
    std::cout << "arguments missing" << std::endl;
    exit(1);
  }

  std::string outfile_path{ argv[1 + argshift] };

  //  -----   test if file exists   -------------------------------------
  if (!overwrite) {
    bool outfile_exists = false;
    {	// scope for limited lifetime of ifstream object
      std::ifstream outf_read{ outfile_path };
      if (outf_read.is_open())
        outfile_exists = true;
      outf_read.close();
      if (outfile_exists) {
        std::cout << "Specified output file exists. Aborting." << std::endl;
        exit(2);
      }
    }
  }
  // ----------------------------------------------------------------------
  std::ofstream outf_write{ outfile_path };
  if (!outf_write.is_open()) {
    std::cout << "Specified output file cannot be opened. Aborting."
              << std::endl;
    exit(3);
  }
  else
    std::cout << "# Event data will be written to : " << outfile_path
              << std::endl;

  // ----------------------------------------------------------------------
  long acqtime_ms_local = 0;
  //bool acqtime_ms_ok = false;
  {
    std::istringstream iss{ argv[2 + argshift] };
    iss >> acqtime_ms_local;
    if (!iss.fail()) {
      //acqtime_ms_ok = true;
      std::cout << "# Acquisition time is " << acqtime_ms_local << " ms"
                << std::endl;
    }
    else {
      std::cout << "Specified acquisition time " << argv[2 + argshift]
                << " cannot be parsed to an integer number" << std::endl;
      outf_write.close();
      exit(4);
    }
  }
  // --------------------------------------------------------------------------
  long number_of_measurements = 0;
  //bool number_of_measurements_ok = false;
  {
    std::istringstream iss(argv[3 + argshift]);
    iss >> number_of_measurements;
    if (!iss.fail()) {
      //number_of_measurements_ok = true;
      std::cout << "# Number of measurements is " << number_of_measurements
                << std::endl;
    }
    else {
      std::cout << "Specified number of measurements " << argv[3 + argshift]
                 << " cannot be parsed to an integer number" << std::endl;
      outf_write.close();
      exit(5);
    }
  }
  // ---------------------------------------------------------------------------

  usercallbacks_example e;
  return e.write_tdc_events_to_textfile(outf_write, acqtime_ms_local,
                                 number_of_measurements);
}
